/***************************************************************************
 $RCSfile: app.cpp,v $
                             -------------------
    cvs         : $Id: app.cpp,v 1.24 2005/03/28 08:48:41 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "app.h"
#include "timport.h"

#include <aqbanking/jobsingletransfer.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/directory.h>
#include <gwenhywfar/text.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#ifdef WIN32
# include <io.h> // for open()
# define strcasecmp stricmp
# define snprintf _snprintf
#endif



App::App(const char *appname, const char *fname)
  :Banking(appname, fname), _lastUniqueId(0) {


}



App::~App(){
}



std::list<Account*> &App::getAccounts(){
  return _accounts;
}



int App::init() {
  GWEN_DB_NODE *db;
  GWEN_DB_NODE *dbT;
  int rv;
  GWEN_TYPE_UINT32 currentVersion;

  rv=Banking::init();
  if (rv)
    return rv;

  /* read configuration */
  db=getAppData();
  db=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "prg");

  currentVersion=
    (AQBANKING_KDE_VERSION_MAJOR<<24)+
    (AQBANKING_KDE_VERSION_MINOR<<16)+
    (AQBANKING_KDE_VERSION_PATCHLEVEL<<8)+
    AQBANKING_KDE_VERSION_BUILD;

  _lastVersion=GWEN_DB_GetIntValue(db, "lastVersion", 0, 0);

  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT,
                       "tMatcherRules");
  if (dbT)
    _xaMatcherRules=GWEN_DB_Group_dup(dbT);
  else
    _xaMatcherRules=GWEN_DB_Group_new("tMatcherRules");

  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT,
                       "exportRules");
  if (dbT)
    _exportRules=GWEN_DB_Group_dup(dbT);
  else
    _exportRules=GWEN_DB_Group_new("exportRules");

  _lastUniqueId=GWEN_DB_GetIntValue(db, "lastUniqueId", 0, 0);
  _lastPayeeId=GWEN_DB_GetIntValue(db, "lastPayeeId", 0, 0);

  _optionAutoAssignPayee=GWEN_DB_GetIntValue(db, "autoAssignPayee", 0, 1);
  _optionAutoAskForPayee=GWEN_DB_GetIntValue(db, "autoAskForPayee", 0, 1);

  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "accounts");
  if (dbT) {
    GWEN_DB_NODE *dbAccount;

    dbAccount=GWEN_DB_FindFirstGroup(dbT, "account");
    while(dbAccount) {
      Account *a;

      a=new Account(this);
      if (!a->fromDb(dbAccount)) {
	DBG_ERROR(0, "Bad account in configuration file");
	delete a;
      }
      else {
	AB_ACCOUNT *ba;

	ba=getAccount(a->getBankingId());
	if (ba) {
	  const char *s;

	  DBG_INFO(0, "Updating account");
	  s=AB_Account_GetBankName(ba);
	  if (s)
	    a->setBankName(s);
	  s=AB_Account_GetAccountName(ba);
	  if (s)
	    a->setAccountName(s);
	  s=AB_Account_GetOwnerName(ba);
	  if (s)
	    a->setOwnerName(s);
        }
        else {
          DBG_NOTICE(0, "Account "GWEN_TYPE_TMPL_UINT32" no longer "
                     "supported by AqBanking",
                     a->getBankingId());
        }
	DBG_INFO(0, "Adding account");
	_accounts.push_back(a);
      }
      dbAccount=GWEN_DB_FindNextGroup(dbAccount, "account");
    } /* while */
  }

  // read payees
  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "payees");
  if (dbT) {
    GWEN_DB_NODE *dbPayee;

    dbPayee=GWEN_DB_FindFirstGroup(dbT, "payee");
    while(dbPayee) {
      Payee *py;

      py=new Payee();
      if (!py->fromDb(dbPayee)) {
        DBG_ERROR(0, "Bad payee in configuration file");
        delete py;
      }
      else {
        _payees.push_back(py);
      }
      dbPayee=GWEN_DB_FindNextGroup(dbPayee, "payee");
    } /* while */
  }

  if (_lastVersion<
      ((0<<24)+
       (9<<16)+
       (18<<8)+
       4)) {
    DBG_NOTICE(0, "Updating from old version, scanning all jobs once");
    /* scan all jobs */
    scanAllJobs();
  }

  /* read transfers */
  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "transfers");
  if (dbT) {
    GWEN_DB_NODE *dbTransfer;

    dbTransfer=GWEN_DB_FindFirstGroup(dbT, "transfer");
    while(dbTransfer) {
      _transfers.push_back(new Transfer(dbTransfer));
      dbTransfer=GWEN_DB_FindNextGroup(dbTransfer, "transfer");
    } // while
  }

  /* sync account list */
  if (!updateAccountList())
    return AB_ERROR_GENERIC;
  return 0;
}



int App::fini() {
  std::list<Account*>::iterator it;
  std::list<Payee*>::iterator pit;
  std::list<RefPointer<Transfer> >::iterator itt;
  GWEN_DB_NODE *db;
  GWEN_DB_NODE *dbT;
  int rv;

  db=getAppData();
  db=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "prg");
  assert(db);

  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "lastVersion",
                      (AQBANKING_KDE_VERSION_MAJOR<<24)+
                      (AQBANKING_KDE_VERSION_MINOR<<16)+
                      (AQBANKING_KDE_VERSION_PATCHLEVEL<<8)+
                      AQBANKING_KDE_VERSION_BUILD);

  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "lastPayeeId", _lastPayeeId);
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
		      "lastUniqueId", _lastUniqueId);

  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "autoAssignPayee",
                      _optionAutoAssignPayee);
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "autoAskForPayee",
                      _optionAutoAskForPayee);

  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "tMatcherRules");
  assert(dbT);
  GWEN_DB_AddGroupChildren(dbT, _xaMatcherRules);
  GWEN_DB_Group_free(_xaMatcherRules);
  _xaMatcherRules=0;

  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "exportRules");
  assert(dbT);
  GWEN_DB_AddGroupChildren(dbT, _exportRules);
  GWEN_DB_Group_free(_exportRules);
  _exportRules=0;

  // save accounts
  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "accounts");
  for (it=_accounts.begin(); it!=_accounts.end(); it++) {
    GWEN_DB_NODE *dbAccount;

    dbAccount=GWEN_DB_GetGroup(dbT, GWEN_PATH_FLAGS_CREATE_GROUP,
			       "account");
    assert(dbAccount);
    if (!(*it)->toDb(dbAccount)) {
      DBG_ERROR(0, "Could not save account");
      return AB_ERROR_GENERIC;
    }
  }

  // save payees
  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "payees");
  for (pit=_payees.begin(); pit!=_payees.end(); pit++) {
    GWEN_DB_NODE *dbPayee;

    dbPayee=GWEN_DB_GetGroup(dbT, GWEN_PATH_FLAGS_CREATE_GROUP,
                             "payee");
    assert(dbPayee);
    if (!(*pit)->toDb(dbPayee)) {
      DBG_ERROR(0, "Could not save payee");
      return AB_ERROR_GENERIC;
    }
  }


  // save transfers
  dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "transfers");
  for (itt=_transfers.begin(); itt!=_transfers.end(); itt++) {
    GWEN_DB_NODE *dbTransfer;

    dbTransfer=GWEN_DB_GetGroup(dbT, GWEN_PATH_FLAGS_CREATE_GROUP,
                                "transfer");
    assert(dbTransfer);
    if (!(*itt).ref().toDb(dbTransfer)) {
      DBG_ERROR(0, "Could not save transfer");
      return AB_ERROR_GENERIC;
    }
  }

  // clear lists
  _accounts.clear();
  _transfers.clear();
  _payees.clear();

  rv=Banking::fini();
  if (rv) {
    return rv;
  }

  return 0;
}



Account *App::findAccount(const char *bankCode,
                          const char *accountNumber) {
  std::list<Account*>::iterator it;

  if (!bankCode)
    bankCode="";
  if (!accountNumber)
    accountNumber=0;

  for (it=_accounts.begin(); it!=_accounts.end(); it++) {
    if ((strcasecmp((*it)->getBankCode().c_str(),
                    bankCode)==0) &&
        (strcasecmp((*it)->getAccountNumber().c_str(),
                    accountNumber)==0))
      return *it;
  }

  return 0;
}



bool App::updateAccountList() {
  std::list<AB_ACCOUNT*> oba;
  std::list<AB_ACCOUNT*>::iterator it;

  oba=Banking::getAccounts();
  if (oba.size()==0) {
    DBG_WARN(0, "No accounts");
  }
  for (it=oba.begin(); it!=oba.end(); it++) {
    Account *a;

    a=findAccount(AB_Account_GetBankCode(*it),
                  AB_Account_GetAccountNumber(*it));
    if (!a) {
      /* account is new, create it */
      a=new Account(this, *it);
      DBG_INFO(0, "Importing account");
      _accounts.push_back(a);
    }
    else {
      DBG_INFO(0, "Account already exists");
      a->setBankingId(AB_Account_GetUniqueId(*it));
    }
  }

  return true;
}



std::list<RefPointer<Transfer> > &App::getTransfers(){
  return _transfers;
}



void App::addTransfer(RefPointer<Transfer> t){
  _transfers.push_back(t);
}



void App::infoTransactionsChanged(Account *a){
  a->infoTransactionsChanged();
}



GWEN_TYPE_UINT32 App::getNextUniqueId() {
  return ++_lastUniqueId;
}



bool App::importContext(AB_IMEXPORTER_CONTEXT *ctx,
                        GWEN_TYPE_UINT32 flags){
  TransactionImporter imp(this, ctx, flags);

  if (!imp.checkTransactions()) {
    DBG_INFO(0, "Error checking transactions");
    return false;
  }

  if (!imp.importContext()) {
    DBG_INFO(0, "Error importing transactions");
    return false;
  }

  return true;
}



RefPointer<Transfer> App::findTransferByTransactionId(GWEN_TYPE_UINT32 id){
  std::list<RefPointer<Transfer> >::iterator it;

  for (it=_transfers.begin(); it!=_transfers.end(); it++) {
    if (id==(*it).ref().getTransactionId())
      return *it;
  } // for

  return 0;
}



void App::scanJobList(AB_JOB_LIST2 *jl) {
  AB_JOB_LIST2_ITERATOR *it;
  AB_JOB *j;

  it=AB_Job_List2_First(jl);
  if (!it)
    return;

  j=AB_Job_List2Iterator_Data(it);
  assert(j);
  while(j) {
    switch(AB_Job_GetType(j)) {
    case AB_Job_TypeGetBalance:
      break;
    case AB_Job_TypeGetTransactions:
      break;
    case AB_Job_TypeTransfer: {
      const AB_TRANSACTION *bt;

      bt=AB_JobSingleTransfer_GetTransaction(j);
      if (bt) {
        GWEN_TYPE_UINT32 tid;
        RefPointer<Transfer> t;

        DBG_INFO(0, "Checking transfer");
        tid=AB_Transaction_GetUniqueId(bt);
        if (tid) {
          t=findTransferByTransactionId(tid);
          if (t.isValid()) {
            DBG_INFO(0, "Transfer is known");
            if (t.ref().getTransactionId()==0)
              t.ref().setTransactionId(getNextUniqueId());
          }
          if (!t.isValid()) {
            t=new Transfer(bt);
            t.ref().setJobId(AB_Job_GetJobId(j));
            DBG_INFO(0, "Adding transfer");
            addTransfer(t);
          }
          if (!t.ref().getDate()) {
            DBG_INFO(0, "Adjusting date");
            t.ref().setDate(AB_Job_GetLastStatusChange(j));
          }
          if (t.ref().getStatus()<Transfer::StatusOk) {
            const GWEN_STRINGLIST *sl;

            sl=AB_Transaction_GetRemoteName(bt);
            if (sl) {
              GWEN_STRINGLISTENTRY *se;

              t.ref().clearRemoteName();
              se=GWEN_StringList_FirstEntry(sl);
              while(se) {
                const char *p;

                p=GWEN_StringListEntry_Data(se);
                assert(p);
                t.ref().addRemoteName(p);
                se=GWEN_StringListEntry_Next(se);
              } /* while */
            }
            sl=AB_Transaction_GetPurpose(bt);
            if (sl) {
              GWEN_STRINGLISTENTRY *se;

              t.ref().clearPurpose();
              se=GWEN_StringList_FirstEntry(sl);
              while(se) {
                const char *p;

                p=GWEN_StringListEntry_Data(se);
                assert(p);
                t.ref().addPurpose(p);
                se=GWEN_StringListEntry_Next(se);
              } /* while */
            }

            switch(AB_Job_GetStatus(j)) {
            case AB_Job_StatusEnqueued:
              DBG_INFO(0, "Status: Enqueued");
              t.ref().setStatus(Transfer::StatusEnqueued);
              break;

            case AB_Job_StatusNew:
            case AB_Job_StatusUpdated:
            case AB_Job_StatusSent:
            case AB_Job_StatusPending:
            case AB_Job_StatusDeferred:
              DBG_INFO(0, "Status: Pending");
              t.ref().setStatus(Transfer::StatusPending);
              break;
            case AB_Job_StatusFinished:
              DBG_INFO(0, "Status: Finished");
              t.ref().setStatus(Transfer::StatusOk);
              break;
            case AB_Job_StatusError:
              DBG_INFO(0, "Status: Error");
              t.ref().setStatus(Transfer::StatusError);
              break;
            default:
              t.ref().setStatus(Transfer::StatusUnknown);
              DBG_INFO(0, "Transfer status unknown");
              break;
            } // switch
          } // if status is changeable
          else {
            DBG_INFO(0, "Unmutable status %d", t.ref().getStatus());
          }
        } // if transaction id is valid
        else {
          DBG_INFO(0, "Ignoring job with no transfer id");
        }
      }
      break;
    }
    case AB_Job_TypeDebitNote:
      break;
    default:
      break;
    } // switch

    j=AB_Job_List2Iterator_Next(it);
  } // while
  AB_Job_List2Iterator_free(it);
}



void App::scanAllJobs() {
  AB_JOB_LIST2 *jl;

  jl=AB_Banking_GetPendingJobs(getCInterface());
  if (jl) {
    scanJobList(jl);
    AB_Job_List2_FreeAll(jl);
  }

  jl=AB_Banking_GetArchivedJobs(getCInterface());
  if (jl) {
    scanJobList(jl);
    AB_Job_List2_FreeAll(jl);
  }

  jl=AB_Banking_GetFinishedJobs(getCInterface());
  if (jl) {
    scanJobList(jl);
    AB_Job_List2_FreeAll(jl);
  }

  jl=AB_Banking_GetEnqueuedJobs(getCInterface());
  if (jl) {
    scanJobList(jl);
    AB_Job_List2_free(jl);
  }
}



GWEN_DB_NODE *App::getTransactionMatcherRules(){
  return _xaMatcherRules;
}



void App::setTransactionMatcherRules(GWEN_DB_NODE *db){
  assert(db);
  GWEN_DB_Group_free(_xaMatcherRules);
  _xaMatcherRules=GWEN_DB_Group_dup(db);
}



GWEN_DB_NODE *App::getExportRules(){
  return _exportRules;
}



void App::setExportRules(GWEN_DB_NODE *db){
  assert(db);
  GWEN_DB_Group_free(_exportRules);
  _exportRules=GWEN_DB_Group_dup(db);
}



bool App::ruleAppliesToTransaction(GWEN_DB_NODE *dbRule,
                                   const Transaction *t) {
  int allAccounts;
  GWEN_DB_NODE *dbData;
  GWEN_DB_NODE *dbT;
  bool accMatch=true;
  bool xaMatch=true;

  dbData=GWEN_DB_GetGroup(dbRule, GWEN_DB_FLAGS_DEFAULT, "data");
  assert(dbData);

  allAccounts=GWEN_DB_GetIntValue(dbData, "account/allAccounts", 0, 0);

  // check date
  dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "date");
  if (dbT) {
    GWEN_TIME *fromTime=0;
    GWEN_TIME *toTime=0;
    int mode;
    GWEN_TIME *cti;
    GWEN_TIME *tti;
    const GWEN_TIME *ti;
    int year, month, day;
    int tyear, tmonth, tday;
    GWEN_TYPE_UINT32 secs;
    const char *s;

    cti=GWEN_CurrentTime();
    assert(cti);
    GWEN_Time_GetBrokenDownDate(cti, &day, &month, &year);

    mode=GWEN_DB_GetIntValue(dbT, "range", 0, 0);
    switch(mode) {
    case TMatcherDateRange_All:
      // nothing to be done in this case
      break;
    case TMatcherDateRange_UntilToday:
      toTime=GWEN_Time_dup(cti);
      break;
    case TMatcherDateRange_CurrentMonth:
      fromTime=GWEN_Time_new(year, month, 1, 0, 0, 0, 0);
      // calculate last day of the current month
      tyear=year;
      tmonth=month+1;
      if (tmonth>11) {
        tmonth-=12;
        tyear++;
      }
      tti=GWEN_Time_new(tyear, tmonth, 1, 0, 0, 0, 0);
      secs=GWEN_Time_Seconds(tti);
      secs-=60*60*24;
      GWEN_Time_free(tti);
      toTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_MonthToDate:
      fromTime=GWEN_Time_new(year, month, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_dup(cti);
      break;
    case TMatcherDateRange_CurrentYear:
      fromTime=GWEN_Time_new(year, 0, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_new(year, 11, 31, 0, 0, 0, 0);
      break;
    case TMatcherDateRange_YearToDate:
      fromTime=GWEN_Time_new(year, 0, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_dup(cti);
      break;
    case TMatcherDateRange_LastMonth:
      tti=GWEN_Time_new(year, month, 1, 0, 0, 0, 0);
      secs=GWEN_Time_Seconds(tti);
      secs-=60*60*24;
      GWEN_Time_free(tti);
      toTime=GWEN_Time_fromSeconds(secs);
      GWEN_Time_GetBrokenDownDate(toTime, &tday, &tmonth, &tyear);
      fromTime=GWEN_Time_new(tyear, tmonth, 1, 0, 0, 0, 0);
      break;
    case TMatcherDateRange_LastYear:
      fromTime=GWEN_Time_new(year-1, 0, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_new(year-1, 11, 31, 0, 0, 0, 0);
      break;
    case TMatcherDateRange_Last30Days:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*30;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_Last3Months:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*30*3;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_Last6Months:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*30*6;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_Last12Months:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*365;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_UserDefined:
      s=GWEN_DB_GetCharValue(dbT, "fromDate", 0, 0);
      if (s) {
        fromTime=GWEN_Time_fromString(s, "YYYY/MM/DD");
        assert(fromTime);
      }

      s=GWEN_DB_GetCharValue(dbT, "toDate", 0, 0);
      if (s) {
        toTime=GWEN_Time_fromString(s, "YYYY/MM/DD");
        assert(toTime);
      }
      break;
    default:
      DBG_ERROR(0, "Unknown date range type (%d)", mode);
      abort();
    } // switch
    GWEN_Time_free(cti);

    // now compare dates
    ti=t->getDate();
    if (!ti)
      ti=t->getValutaDate();
    if (!ti) {
      DBG_ERROR(0, "No date/valutaDate in transaction");
      return false;
    }
    if (fromTime) {
      if (GWEN_Time_Diff(ti, fromTime)<0.0) {
        DBG_VERBOUS(0, "fromTime does not match");
        return false;
      }
    }
    if (toTime) {
      if (GWEN_Time_Diff(toTime, ti)<0.0) {
        DBG_VERBOUS(0, "toTime does not match");
        return false;
      }
    }
    GWEN_Time_free(toTime);
    GWEN_Time_free(fromTime);
  } // if date mentioned

  DBG_VERBOUS(0, "Checking account");
  accMatch=false;
  if (allAccounts!=0)
    accMatch=true;
  else {
    dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                         "account/selected");
    if (dbT) {
      DBG_VERBOUS(0, "Have selected group...");
      dbT=GWEN_DB_FindFirstGroup(dbT, "account");
      while (dbT) {
        const char *accountId;
        const char *bankId;

        DBG_VERBOUS(0, "Have account group...");
        bankId=GWEN_DB_GetCharValue(dbT, "bankId", 0, "");
        accountId=GWEN_DB_GetCharValue(dbT, "accountId", 0, "");

        DBG_VERBOUS(0, "BankID: %s, AccountID: %s",
                   bankId, accountId);
        if (strcasecmp(t->getLocalBankCode().c_str(),
                       bankId)==0 &&
            strcasecmp(t->getLocalAccountNumber().c_str(),
                       accountId)==0) {
          accMatch=true;
          break;
        }
        dbT=GWEN_DB_FindNextGroup(dbT, "account");
      } // while
    } // if "account/selected"
  }

  if (!accMatch) {
    DBG_VERBOUS(0, "Account does not match");
    return false;
  }

  xaMatch=true;

  // check purpose
  dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "text");
  if (dbT) {
    const char *s;

    s=GWEN_DB_GetCharValue(dbT, "purpose", 0, 0);
    if (s) {
      std::string purpose;
      std::list<std::string>::const_iterator sit;
      int senseCase;

      senseCase=GWEN_DB_GetIntValue(dbT, "purposeSenseCase", 0, 0);

      /* combine purpose */
      for (sit=t->getPurpose().begin();
           sit!=t->getPurpose().end();
           sit++) {
        if (!purpose.empty() && !(*sit).empty())
          purpose+=" ";
        purpose+=*sit;
      }
      if (-1==GWEN_Text_ComparePattern(purpose.c_str(),
                                       s,
                                       senseCase)) {
        DBG_VERBOUS(0, "purpose does not match");
        xaMatch=false;
      }
    }
  }

  if (!xaMatch)
    return false;

  // check amount
  dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "amount");
  if (dbT) {

  }

  // check payee
  dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "payee");
  if (dbT) {
    const char *s;

    s=GWEN_DB_GetCharValue(dbT, "payee", 0, 0);
    if (s) {
      std::string payee;
      std::list<std::string>::const_iterator sit;
      int senseCase;

      senseCase=GWEN_DB_GetIntValue(dbT, "payeeSenseCase", 0, 0);

      /* combine purpose */
      for (sit=t->getRemoteName().begin();
           sit!=t->getRemoteName().end();
           sit++) {
        if (!payee.empty() && !(*sit).empty())
          payee+=" ";
        payee+=*sit;
      }
      if (-1==GWEN_Text_ComparePattern(payee.c_str(),
                                       s,
                                       senseCase)) {
        DBG_VERBOUS(0, "payee does not match ([%s] != [%s]",
                 payee.c_str(), s);
        xaMatch=false;
      }
    }
  }

  if (!xaMatch)
    return false;

  return true;
}



bool App::getMatchingTransactions(GWEN_DB_NODE *dbRule,
                                  std::list<RefPointer<Transaction> > &tl){
  std::list<Account*>::iterator ait;
  int allAccounts;
  GWEN_TIME *fromTime;
  GWEN_TIME *toTime;
  GWEN_DB_NODE *dbData;
  GWEN_DB_NODE *dbT;

  fromTime=toTime=0;

  dbData=GWEN_DB_GetGroup(dbRule, GWEN_DB_FLAGS_DEFAULT, "data");
  assert(dbData);

  allAccounts=GWEN_DB_GetIntValue(dbData, "account/allAccounts", 0, 0);

  // check date
  dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "date");
  if (dbT) {
    int mode;
    GWEN_TIME *cti;
    GWEN_TIME *tti;
    int year, month, day;
    int tyear, tmonth, tday;
    GWEN_TYPE_UINT32 secs;
    const char *s;

    cti=GWEN_CurrentTime();
    assert(cti);
    GWEN_Time_GetBrokenDownDate(cti, &day, &month, &year);

    mode=GWEN_DB_GetIntValue(dbT, "range", 0, 0);
    switch(mode) {
    case TMatcherDateRange_All:
      // nothing to be done in this case
      break;
    case TMatcherDateRange_UntilToday:
      toTime=GWEN_Time_dup(cti);
      break;
    case TMatcherDateRange_CurrentMonth:
      fromTime=GWEN_Time_new(year, month, 1, 0, 0, 0, 0);
      // calculate last day of the current month
      tyear=year;
      tmonth=month+1;
      if (tmonth>11) {
        tmonth-=12;
        tyear++;
      }
      tti=GWEN_Time_new(tyear, tmonth, 1, 0, 0, 0, 0);
      secs=GWEN_Time_Seconds(tti);
      secs-=60*60*24;
      GWEN_Time_free(tti);
      toTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_MonthToDate:
      fromTime=GWEN_Time_new(year, month, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_dup(cti);
      break;
    case TMatcherDateRange_CurrentYear:
      fromTime=GWEN_Time_new(year, 0, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_new(year, 11, 31, 0, 0, 0, 0);
      break;
    case TMatcherDateRange_YearToDate:
      fromTime=GWEN_Time_new(year, 0, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_dup(cti);
      break;
    case TMatcherDateRange_LastMonth:
      tti=GWEN_Time_new(year, month, 1, 0, 0, 0, 0);
      secs=GWEN_Time_Seconds(tti);
      secs-=60*60*24;
      GWEN_Time_free(tti);
      toTime=GWEN_Time_fromSeconds(secs);
      GWEN_Time_GetBrokenDownDate(toTime, &tday, &tmonth, &tyear);
      fromTime=GWEN_Time_new(tyear, tmonth, 1, 0, 0, 0, 0);
      break;
    case TMatcherDateRange_LastYear:
      fromTime=GWEN_Time_new(year-1, 0, 1, 0, 0, 0, 0);
      toTime=GWEN_Time_new(year-1, 11, 31, 0, 0, 0, 0);
      break;
    case TMatcherDateRange_Last30Days:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*30;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_Last3Months:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*30*3;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_Last6Months:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*30*6;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_Last12Months:
      secs=GWEN_Time_Seconds(cti);
      secs-=60*60*24*365;
      fromTime=GWEN_Time_fromSeconds(secs);
      break;
    case TMatcherDateRange_UserDefined:
      s=GWEN_DB_GetCharValue(dbT, "fromDate", 0, 0);
      if (s) {
        fromTime=GWEN_Time_fromString(s, "YYYY/MM/DD");
        assert(fromTime);
      }

      s=GWEN_DB_GetCharValue(dbT, "toDate", 0, 0);
      if (s) {
        toTime=GWEN_Time_fromString(s, "YYYY/MM/DD");
        assert(toTime);
      }
      break;
    default:
      DBG_ERROR(0, "Unknown date range type (%d)", mode);
      abort();
    } // switch
    GWEN_Time_free(cti);
  } // if date mentioned


  // read all matching transactions
  for (ait=_accounts.begin(); ait!=_accounts.end(); ait++) {
    bool accMatch;

    DBG_NOTICE(0, "Checking account");
    accMatch=false;
    if (allAccounts!=0)
      accMatch=true;
    else {
      dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                           "account/selected");
      if (dbT) {
        DBG_NOTICE(0, "Have selected group...");
        dbT=GWEN_DB_FindFirstGroup(dbT, "account");
        while (dbT) {
          const char *accountId;
          const char *bankId;

          DBG_NOTICE(0, "Have account group...");
          bankId=GWEN_DB_GetCharValue(dbT, "bankId", 0, "");
          accountId=GWEN_DB_GetCharValue(dbT, "accountId", 0, "");

          DBG_NOTICE(0, "BankID: %s, AccountID: %s",
                     bankId, accountId);
          if (strcasecmp((*ait)->getBankCode().c_str(),
                         bankId)==0 &&
              strcasecmp((*ait)->getAccountNumber().c_str(),
                         accountId)==0) {
            accMatch=true;
            break;
          }
          dbT=GWEN_DB_FindNextGroup(dbT, "account");
        } // while
      } // if "account/selected"
    }

    if (accMatch) {
      std::list<RefPointer<Transaction> > tmptl;
      std::list<RefPointer<Transaction> >::iterator tit;

      DBG_NOTICE(0, "Account matches");
      if (1) {
        GWEN_BUFFER *tbuf;

        tbuf=GWEN_Buffer_new(0, 256, 0, 1);
        if (fromTime) {
          GWEN_Time_toString(fromTime, "YYYY/MM/DD", tbuf);
          DBG_NOTICE(0, "FromTime: %s", GWEN_Buffer_GetStart(tbuf));
          GWEN_Buffer_Reset(tbuf);
        }
        if (toTime) {
          GWEN_Time_toString(toTime, "YYYY/MM/DD", tbuf);
          DBG_NOTICE(0, "ToTime: %s", GWEN_Buffer_GetStart(tbuf));
          GWEN_Buffer_Reset(tbuf);
        }
        GWEN_Buffer_free(tbuf);
      }

      // load all transactions of this account
      DBG_NOTICE(0, "Loading transactions");
      if (!(*ait)->loadTransactions(fromTime, toTime, tmptl)) {
        DBG_ERROR(0, "Error loading transactions");
        GWEN_Time_free(fromTime);
        GWEN_Time_free(toTime);
        return false;
      }

      // copy matching transactions to final list
      DBG_NOTICE(0, "Copying transactions (%d)", tmptl.size());
      for (tit=tmptl.begin(); tit!=tmptl.end(); tit++) {
	bool xaMatch;

        xaMatch=true;

        DBG_DEBUG(0, "Checking transactions");
        // check purpose
        dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                             "text");
	if (dbT) {
	  const char *s;

	  s=GWEN_DB_GetCharValue(dbT, "purpose", 0, 0);
	  if (s) {
	    std::string purpose;
	    std::list<std::string>::const_iterator sit;
	    int senseCase;

	    senseCase=GWEN_DB_GetIntValue(dbT, "purposeSenseCase", 0, 0);

            /* combine purpose */
	    for (sit=(*tit).ref().getPurpose().begin();
		 sit!=(*tit).ref().getPurpose().end();
		 sit++) {
	      if (!purpose.empty() && !(*sit).empty())
		purpose+=" ";
	      purpose+=*sit;
	    }
	    if (-1==GWEN_Text_ComparePattern(purpose.c_str(),
					     s,
					     senseCase))
	      xaMatch=false;
	  }
	}

        // check amount
        dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                             "amount");
        if (dbT) {

        }

        // check payee
        dbT=GWEN_DB_GetGroup(dbData, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                             "payee");
        if (dbT) {
	  const char *s;

	  s=GWEN_DB_GetCharValue(dbT, "payee", 0, 0);
	  if (s) {
	    std::string payee;
	    std::list<std::string>::const_iterator sit;
	    int senseCase;

	    senseCase=GWEN_DB_GetIntValue(dbT, "payeeSenseCase", 0, 0);

            /* combine purpose */
	    for (sit=(*tit).ref().getRemoteName().begin();
		 sit!=(*tit).ref().getRemoteName().end();
		 sit++) {
	      if (!payee.empty() && !(*sit).empty())
		payee+=" ";
	      payee+=*sit;
	    }
	    if (-1==GWEN_Text_ComparePattern(payee.c_str(),
					     s,
					     senseCase))
	      xaMatch=false;
	  }

        }


        if (xaMatch) {
          DBG_NOTICE(0, "Adding transaction");
          tl.push_back(*tit);
        }
      } // for
    } // if account matches
  } // for

  GWEN_Time_free(fromTime);
  GWEN_Time_free(toTime);
  return true;
}



int App::exportTransactions(GWEN_DB_NODE *dbRule,
                            const std::list<RefPointer<Transaction> > &tl){
  AB_IMEXPORTER_CONTEXT *ctx;
  std::list<RefPointer<Transaction> >::const_iterator it;
  AB_IMEXPORTER *exporter;
  const char *exporterName;
  GWEN_DB_NODE *dbProfile;
  int rv;
  GWEN_BUFFEREDIO *bio;
  const char *fname;
  int fd;
  GWEN_ERRORCODE err;
  GWEN_TYPE_UINT32 wid;

  /* get exporter name */
  exporterName=GWEN_DB_GetCharValue(dbRule, "exporterText", 0, 0);
  if (!exporterName) {
    DBG_ERROR(0, "No exporter set");
    return AB_ERROR_INVALID;
  }

  /* get profile */
  dbProfile=GWEN_DB_GetGroup(dbRule, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                             "profile");
  if (!dbProfile) {
    DBG_ERROR(0, "No export profile set");
    return AB_ERROR_INVALID;
  }

  /* get exporter */
  exporter=AB_Banking_GetImExporter(getCInterface(), exporterName);
  if (!exporter) {
    DBG_ERROR(0, "Exporter %s not found", exporterName);
    return AB_ERROR_INVALID;
  }

  /* check for file */
  fname=GWEN_DB_GetCharValue(dbRule, "outfileText", 0, 0);
  if (!fname) {
    DBG_ERROR(0, "No file name set");
    return AB_ERROR_INVALID;
  }

  /* open file */
  fd=open(fname,
          O_RDWR | O_TRUNC | O_CREAT,
#ifdef WIN32
		  0
#else
		  S_IRUSR | S_IWUSR
#endif
		  );
  if (fd==-1) {
    DBG_ERROR(0, "open(%s): %s", fname, strerror(errno));
    return AB_ERROR_GENERIC;
  }
  bio=GWEN_BufferedIO_File_new(fd);
  GWEN_BufferedIO_SetWriteBuffer(bio, 0, 1024);

  /* add transactions to context */
  ctx=AB_ImExporterContext_new();
  for (it=tl.begin(); it!=tl.end(); it++) {
    AB_TRANSACTION *bt;

    bt=(*it).ref().toBankingTransaction();
    assert(bt);
    AB_ImExporterContext_AddTransaction(ctx, bt);
  } // for

  /* let exporter work */
  wid=AB_Banking_ProgressStart(getCInterface(),
                               "Exporting Transactions ...",
                               "The selected transactions are now being "
                               "exported, please wait...",
                               tl.size());
  rv=AB_ImExporter_Export(exporter, ctx, bio, dbProfile);
  AB_ImExporterContext_free(ctx);
  if (rv) {
    AB_Banking_ProgressLog(getCInterface(),
                           wid,
                           AB_Banking_LogLevelError,
                           "Export plugin signalled an error");
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
  }
  else {
    err=GWEN_BufferedIO_Close(bio);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      AB_Banking_ProgressLog(getCInterface(),
                             wid,
                             AB_Banking_LogLevelError,
                             "Export plugin signalled an error");
      GWEN_BufferedIO_Abandon(bio);
      GWEN_BufferedIO_free(bio);
      return AB_ERROR_GENERIC;
    }
    AB_Banking_ProgressLog(getCInterface(),
                           wid,
                           AB_Banking_LogLevelNotice,
                           "Export completed.");
    GWEN_BufferedIO_free(bio);
  }
  AB_Banking_ProgressEnd(getCInterface(), wid);
  return rv;
}



Payee *App::findPayeeById(const char *id){
  std::list<Payee*>::iterator it;

  assert(id);
  assert(*id);
  for (it=_payees.begin(); it!=_payees.end(); it++) {
    if (strcasecmp(id, (*it)->id().c_str())==0)
      return *it;
  }

  return 0;
}



Payee *App::findPayeeByName(const char *name){
  std::list<Payee*>::iterator it;

  for (it=_payees.begin(); it!=_payees.end(); it++) {
    if (-1!=GWEN_Text_ComparePattern((*it)->name().c_str(), name, 0))
      return *it;
  }
  return 0;
}



Payee *App::findPayeeByTransactionMatch(const Transaction *t){
  std::list<Payee*>::iterator it;

  for (it=_payees.begin(); it!=_payees.end(); it++) {
    GWEN_DB_NODE *dbRule;

    dbRule=GWEN_DB_FindFirstGroup((*it)->rules(), "rule");
    while(dbRule) {
      if (ruleAppliesToTransaction(dbRule, t))
        return *it;
      dbRule=GWEN_DB_FindNextGroup(dbRule, "rule");
    }
  } // for payees
  return 0;
}



const std::list<Payee*> &App::getPayees(){
  return _payees;
}



bool App::addPayee(Payee *p){
  char numbuf[32];
  Payee *tp;
  int i;

  tp=0;
  for (i=0; i<100; i++) {

    snprintf(numbuf, sizeof(numbuf), "%d", ++_lastPayeeId);
    tp=findPayeeById(numbuf);
    if (!tp)
      break;
  }
  if (tp) {
    DBG_ERROR(0, "Could not add payee");
    return false;
  }

  p->setId(numbuf);
  _payees.push_back(p);
  return true;
}



bool App::removePayee(Payee *p){
  std::list<Payee*>::iterator it;

  for (it=_payees.begin(); it!=_payees.end(); it++) {
    if (*it==p) {
      _payees.erase(it);
      return true;
    }
  }

  return false;
}



bool App::optionAutoAssignPayee(){
  return _optionAutoAssignPayee;
}



void App::setOptionAutoAssignPayee(bool b){
  _optionAutoAssignPayee=b;
}



bool App::optionAutoAskForPayee(){
  return _optionAutoAskForPayee;
}



void App::setOptionAutoAskForPayee(bool b){
  _optionAutoAskForPayee=b;
}



GWEN_TIME *App::adjustedDate(const GWEN_TIME *pt){
  int hour, minute, sec;

  if (pt) {
    GWEN_Time_GetBrokenDownUtcTime(pt, &hour, &minute, &sec);
    if (hour!=12) {
      int day, month, year;

      GWEN_Time_GetBrokenDownUtcDate(pt, &day, &month, &year);
      if (hour==22 || hour==23) {
        GWEN_TIME *nt;

        nt=GWEN_Time_fromSeconds(GWEN_Time_Seconds(pt)+(24*60*60));
        GWEN_Time_GetBrokenDownUtcDate(nt, &day, &month, &year);
        GWEN_Time_free(nt);
      }

      return GWEN_Time_new(year, month, day, 12, 0, 0, 1);
    }
    return GWEN_Time_dup(pt);
  }

  return 0;
}



GWEN_TYPE_UINT32 App::getLastVersion() const {
  return _lastVersion;
}













/* the following algorithm has been copied and adapted from fuzzy.c of
 * gtk-gnutella (Copyright (c) 2002, Vidar Madsen).
 * Gtk-Gnutella is licensed under the GPL.
 */

GWEN_TYPE_UINT32 App__word_similarity(const char *a, const char *b){
  GWEN_TYPE_UINT32 score = 0;
  size_t l = 0;

  while (*a && *b) {
    if (*a == *b)
      score += 1 << APP_FUZZY_SHIFT;
    else if (*a == b[1]) {
      score += 1 << (APP_FUZZY_SHIFT-2); b++;
    }
    else if (a[1] == *b) {
      score += 1 << (APP_FUZZY_SHIFT-2); a++; l++;
    }
    a++;
    b++;
    l++;
  }

  if ('\0' != *a)
    l += strlen(a);

  return score / l;
}



int App__is_ascii_alnum(unsigned char c) {
  return ((c>='A' || c<='Z') ||
	  (c>='a' || c<='z') ||
	  (c>='0' || c<='9') ||
	  c>=128);
}



GWEN_STRINGLIST *App__make_word_list(const char *str) {
  GWEN_STRINGLIST *sl;

  sl=GWEN_StringList_new();
  while(*str) {
    const char *p;
    size_t size;

    while (*str && !App__is_ascii_alnum(*str))
      str++;
    p=str;
    while (*str && App__is_ascii_alnum(*str)) {
      str++;
    }

    size=(str-p)+1;
    if (*p) {
      char *wptr;
      char *t;
      size_t i;

      wptr=(char*)malloc(size);
      t=wptr;
      for (i=0; i<size; i++)
	*(t++)=tolower(*(p++));
      *t=0;
      GWEN_StringList_AppendString(sl, wptr, 1, 0); /* take this string */
    }
  } /* while */

  return sl;
}



GWEN_TYPE_UINT32 App__cmp_word_list(const char *s,
					  GWEN_STRINGLIST *words) {
  GWEN_TYPE_UINT32 score = 0;
  GWEN_TYPE_UINT32 n = 0;
  GWEN_STRINGLISTENTRY *se;

  se=GWEN_StringList_FirstEntry(words);
  while(se) {
    const char *p;

    n++;
    p=GWEN_StringListEntry_Data(se);
    assert(p);

    if (strcmp(s, p)==0)
      return 1 << APP_FUZZY_SHIFT;
    else
      score+=App__word_similarity(s, p);
    se=GWEN_StringListEntry_Next(se);
  }

  return (n>0)?(score/n):0;
}



GWEN_TYPE_UINT32 App__find_score(GWEN_STRINGLIST *a,
				       GWEN_STRINGLIST *b) {
  GWEN_TYPE_UINT32 score = 0;
  GWEN_TYPE_UINT32 n = 0;
  GWEN_STRINGLISTENTRY *se;

  se=GWEN_StringList_FirstEntry(a);
  while(se) {
    const char *p;

    n++;
    p=GWEN_StringListEntry_Data(se);
    assert(p);

    score+=App__cmp_word_list(p, b);

    se=GWEN_StringListEntry_Next(se);
  }

  return (n>0)?(score/n):0;
}



GWEN_TYPE_UINT32 App_FuzzyCompare(const char *str1, const char *str2) {
  GWEN_STRINGLIST *a;
  GWEN_STRINGLIST *b;
  GWEN_TYPE_UINT32 score;

  a=App__make_word_list(str1);
  b=App__make_word_list(str2);

  score=(App__find_score(a, b) + App__find_score(b, a))/2;

  GWEN_StringList_free(b);
  GWEN_StringList_free(a);

  return score;
}






